/// This module handles all of the additional functions for generating filenames, adminkeys and calculating expiry.
/// Basically stuff that shouldn't be in main.rs, but also doesn't fit in anywhere else.
use crate::{db, CONFIG};
use nanoid::nanoid;
use sqlx::{Pool, Sqlite};
use std::{
    fs,
    path::{Path, PathBuf},
    time::SystemTime,
};

/// Generate a random name for the uploaded file.
/// Files with no file extension are completely randomised
/// Files with 1 or more extensions are split into the filename and extension, and the filename is randomised.
pub async fn generate_filename<S>(length: usize, filename: S) -> PathBuf
where
    S: AsRef<Path>,
{
    // Grab the file extension
    let ext = filename.as_ref().extension();

    // Randomly generate the filename.
    let random_filename: PathBuf = PathBuf::from(nanoid!(length));

    // Add the extension to the new filename, if it exists.
    if let Some(ext) = ext {
        tracing::debug!("generate_filename(File Extension ({:?}) detected)", ext);
        let random_filename: PathBuf = (nanoid!(length) + "." + ext.to_str().unwrap_or("")).into();
        tracing::debug!("generate_filename(New filename: {:?} )", random_filename);
        return random_filename;
    }
    tracing::debug!("generate_filename(New filename: {:?} )", random_filename);
    random_filename
}

// Generate a random admin for the uploaded file
// This should check if the adminkey exists for another file, and regenerate if it does.
pub async fn generate_adminkey(sqlconn: &Pool<Sqlite>) -> String {
    let adminkey_size: usize = CONFIG
        .get_int("operations.adminkey_size")
        .expect("Couldn't find 'adminkey_size' in config. :(")
        .try_into()
        .expect("Problem with converting adminkey_size to usize");
    let adminkey = nanoid!(adminkey_size);
    let result = db::check_adminkey(sqlconn, &adminkey).await;
    if result.is_ok() {
        // Adminkey resolves to a filename, so we better regenerate it.
        return nanoid!(adminkey_size);
    }
    adminkey
}

pub async fn calculate_expiry<S>(sqlconn: &Pool<Sqlite>, filename: S, filesize: i32) -> i32
where
    S: AsRef<str>,
{
    let engine_mode: i64 = CONFIG
        .get_int("operations.engine_mode")
        .expect("Couldn't find 'engine_mode' in config. :(");

    let expiry = match engine_mode {
        1 => engine_1(sqlconn, filename.as_ref()).await,
        2 => engine_2(filesize).await,
        3 => engine_3(sqlconn, filename.as_ref(), filesize).await,
        _ => {
            tracing::error!("Unknown engine mode: {}", engine_mode);
            std::process::exit(2);
        }
    };

    // REVIEW: there MUST be a less intrusive way to log. Without this the entire function could be just the match statement and no `let`s
    // Especially since many of these are just the fn name and arguments passed. There's probably some crate that lets you just #[derive(Log)] or something
    // audron: for the calculate_expiry() function logging https://docs.rs/tracing/latest/tracing/attr.instrument.html might help, tho idk rn if it also logs return value on function exit
    // Volkor note: I tried adding instrument but it didn't return anything, so idk, I've cleaned this up a little bit more, but I'll keep wasteful logging on the mind when I do it.
    tracing::debug!(
        "calculate_expiry(file: {}, expiry: {}) ",
        filename.as_ref(),
        expiry
    );
    expiry
}

// Engine 1 deletes the file depending on the last view
pub async fn engine_1<S>(sqlconn: &Pool<Sqlite>, filename: S) -> i32
where
    S: AsRef<str>,
{
    // Read the list of unsafe mimetypes from the config
    let file_expiry_min: i32 = CONFIG
        .get_int("operations.file_expiry_min")
        .expect("Couldn't find 'file_expiry_min' in config. :(")
        .try_into()
        .expect("The file_expiry_min is wayyyy to big.");

    // Get the last time the file was viewed (in unix time)
    let accesstime_result = db::get_accesss_time(sqlconn, filename.as_ref()).await;

    let accesstime = match accesstime_result {
        Ok(accesstime) => accesstime as i32,
        Err(_) => {
            // If result is an error, we'll want to overwrite whatever casued the broken timestamp.
            SystemTime::now()
                .duration_since(SystemTime::UNIX_EPOCH)
                .unwrap()
                .as_secs() as i32
        }
    };

    // Now we can add expiry_seconds to accesstime, to get 7 days in the future!
    accesstime + file_expiry_min
}

/// Engine 2 deletes files depending on how big the filesize is.
async fn engine_2(filesize: i32) -> i32 {
    // audron REVIEW: also, probably make that engine match use enums and give them descriptive names

    // Load settings from config
    let file_expiry_min: f64 = CONFIG
        .get_float("operations.file_expiry_min")
        .expect("Couldn't find 'file_expiry_min' in config. :(");
    let file_expiry_max: f64 = CONFIG
        .get_float("operations.file_expiry_max")
        .expect("Couldn't find 'file_expiry_min' in config. :(");
    let max_filesize: f64 = CONFIG
        .get_float("operations.max_filesize")
        .expect("Couldn't find 'max_filesize' in config. :(");

    let time = SystemTime::now()
        .duration_since(SystemTime::UNIX_EPOCH)
        .unwrap()
        .as_secs() as i32;

    // Add the lifetime to the current unix time.
    let expiry = time as f64
        + lifetime_0x0(
            file_expiry_min,
            file_expiry_max,
            max_filesize,
            filesize as f64,
        )
        .await;

    expiry as i32
}

async fn engine_3<S>(sqlconn: &Pool<Sqlite>, filename: S, filesize: i32) -> i32
where
    S: AsRef<str>,
{
    todo!()
}

/// This is a quick one that just runs mia's 0x0.st lifetime algorithm
/// This is split from engine_2 so engine_3 can call it too
/// also so we can test it easier.
pub async fn lifetime_0x0(
    file_expiry_min: f64,
    file_expiry_max: f64,
    max_filesize: f64,
    filesize: f64,
) -> f64 {
    file_expiry_min
        + (-file_expiry_max + file_expiry_min) * (filesize / max_filesize - 1.0).powf(3.0)
}

/// Deletes the file with the given name.
pub async fn delete_file<S>(file: S) -> Result<(), std::io::Error>
where
    S: AsRef<Path>,
{
    // Set Destination
    let files_dir = Path::new("files");
    let dest = files_dir.join(file);

    // Attempt to remove the file, or pass the error back up
    fs::remove_file(dest)?;
    Ok(())
}
