#[cfg(test)]
mod tests {
    use crate::{
        db, db_tests,
        engine::{self, generate_filename, lifetime_0x0},
    };
    use std::{collections::HashSet, path::PathBuf};

    // This test checks if a uploaded filename is the correct length. (with a bit added for the extension)
    #[tokio::test]
    async fn test_generate_filename_with_extension_length() {
        let filename_size = 10;
        let filename = PathBuf::from("test.txt");
        let result = generate_filename(filename_size, filename.clone()).await;

        // Compare the length of the generated filename with the expected length of said filename (plus extension)
        assert_eq!(result.to_str().unwrap().len(), filename_size + 4);
        // Check if the generated filename includes the extension
        assert_eq!(result.extension().unwrap(), "txt");
    }

    // This test checks if a uploaded filename is the correct length. (without a extension)
    #[tokio::test]
    async fn test_generate_filename_without_extension_length() {
        let filename_size = 10;
        let filename = PathBuf::from("test");
        let result = generate_filename(filename_size, filename.clone()).await;
        // Compare the length of the generated filename with the expected length of said filename
        assert_eq!(result.to_str().unwrap().len(), filename_size);
    }

    // This test checks if a uploaded file with an extension is actually randomised, or just returned back.
    #[tokio::test]
    async fn test_generate_filename_with_extension_randomness() {
        let filename_size = 10;
        let filename = String::from("test.txt");
        let num_generations = 100;

        let mut filenames = Vec::new();

        for _ in 0..num_generations {
            let new_filename = generate_filename(filename_size, filename.clone()).await;
            filenames.push(new_filename);
        }

        // Ensure all filenames are unique
        let unique_filenames: HashSet<_> = filenames.iter().collect();
        assert_eq!(unique_filenames.len(), num_generations);
    }

    // This test checks if a uploaded file with an extension is actually randomised, or just returned back.
    #[tokio::test]
    async fn test_generate_filename_without_extension_randomness() {
        let length = 10;
        let filename = String::from("test");
        let num_generations = 100;

        let mut filenames = Vec::new();

        for _ in 0..num_generations {
            let new_filename = generate_filename(length, filename.clone()).await;
            filenames.push(new_filename);
        }

        // Ensure all filenames are unique
        let unique_filenames: HashSet<_> = filenames.iter().collect();
        assert_eq!(unique_filenames.len(), num_generations);
    }

    // This test checks if a uploaded file's expiry is being calculated correctly.
    #[tokio::test]
    async fn test_lifetime_0x0() {
        let file_expiry_min: f64 = 604800.0;
        let file_expiry_max: f64 = 31536000.0;
        let max_filesize: f64 = 1073741824.0;

        // This is exessive but I was using this to graph stuff so I'll leave it in.
        let fivebits = lifetime_0x0(file_expiry_min, file_expiry_max, max_filesize, 0.625);
        let fivebytes = lifetime_0x0(file_expiry_min, file_expiry_max, max_filesize, 5.0);
        let fivekb = lifetime_0x0(file_expiry_min, file_expiry_max, max_filesize, 5000.0);
        let fivemb = lifetime_0x0(file_expiry_min, file_expiry_max, max_filesize, 5000000.0);
        let fiftymb = lifetime_0x0(file_expiry_min, file_expiry_max, max_filesize, 50000000.0);
        let fivehundymb = lifetime_0x0(file_expiry_min, file_expiry_max, max_filesize, 500000000.0);
        let sixhundymb = lifetime_0x0(file_expiry_min, file_expiry_max, max_filesize, 600000000.0);
        let sevenhundymb =
            lifetime_0x0(file_expiry_min, file_expiry_max, max_filesize, 700000000.0);
        let eighthundymb =
            lifetime_0x0(file_expiry_min, file_expiry_max, max_filesize, 800000000.0);
        let ninehundymb = lifetime_0x0(file_expiry_min, file_expiry_max, max_filesize, 900000000.0);
        assert_eq!(fivebits.await, 31535999.945987016); // 364.99999937484972179 days
        assert_eq!(fivebytes.await, 31535999.56789613); // 364.99999499879783116 days
        assert_eq!(fivekb.await, 31535567.898139838); // 364.9949988210629499 days
        assert_eq!(fivemb.await, 31105905.144902013); // 360.02205028821776978 days
        assert_eq!(fiftymb.await, 27413052.06938632); // 317.28069524752680763 days
        assert_eq!(fivehundymb.await, 5323764.26538951); // 61.61764196052673981 days
        assert_eq!(sixhundymb.await, 3261376.5997360833); // 37.747414348797256878 days
        assert_eq!(sevenhundymb.await, 1909204.4198811946); // 22.097273378254566722 days
        assert_eq!(eighthundymb.await, 1117331.5795572645); // 12.932078467097968044 days
        assert_eq!(ninehundymb.await, 735841.932496714); // 8.516689033526782637 days
    }

    #[tokio::test]
    async fn test_engine_1() {
        // Set up an in-memory SQLite database for testing
        // ChatGPT did this for me, the whole idea of having the db entirely in memory is amazing for tests!
        let pool = match db_tests::tests::init_test_db().await {
            Ok(pool) => pool,
            Err(err) => {
                tracing::error!("Error initialising database connection:  {:?}", err);
                return;
            }
        };

        // Insert a record for the file's access time using the `add_file()` function
        let result = db::add_file(
            &pool,
            "test.txt",
            "text/plain",
            0, // Expiry
            "adminkey",
            1616601600, // Accessed
            1024,       // filesize
            "127.0.0.1",
            "example.com",
        )
        .await;

        // Check if there was an error with the database insertion
        if let Err(e) = result {
            panic!("Error inserting record into database: {:?}", e);
        }

        // Call the `engine_1()` function to calculate the expiry time
        let expiry_time = engine::engine_1(&pool, "test.txt").await;

        // Verify that the expiry time is 7 days after the access time
        assert_eq!(expiry_time, 1617206400);
    }
}
