use salvo::{
    handler,
    hyper::header::{self, HOST},
    prelude::StatusCode,
    writer::Text,
    Request, Response,
};

use crate::{
    db, engine,
    handlers::{render_template, TemplateStruct},
    SQLITE,
};

/// This takes the adminkey in, and deletes the file that matches it in the DB.
#[handler]
pub async fn delete_file(req: &mut Request, res: &mut Response) {
    // Set up all the things we need for the function later.
    let headers = req.headers().clone();
    let sqlconn = SQLITE.get().unwrap();
    let adminkey: &String = &req.param("adminkey").unwrap();
    tracing::debug!("delete_file(adminkey): {:?}", adminkey);

    // Returns true if the "source" field in the HTTP request's form data is "web", false otherwise or if an error occurs.
    // Since we don't have a 'source' field in the request, we guess from the user agent. It's not perfect, but it works. and it's not really a security problem.
    // Check if the request came from a web browser or something else.
    let user_agent = headers
        .get(header::USER_AGENT)
        .and_then(|value| value.to_str().ok())
        .unwrap_or("");
    let upload_by_web = user_agent.contains("Mozilla");

    // Sends the adminkey to the db and gets the filename to delete back.
    let adminkey_result = db::check_adminkey(sqlconn, adminkey).await;
    // Handle any potential errors.
    match adminkey_result {
        // adminkey returned a filename
        Ok(fname) => {
            // Actually delete the file from disk
            db::delete_file(sqlconn, &fname)
                .await
                .unwrap_or_else(|err| {
                    tracing::error!(
                        "Failed to delete file from the database: {} error: {:?}",
                        &fname,
                        err
                    );
                    0
                });

            engine::delete_file(&fname).await.unwrap_or_else(|err| {
                tracing::error!(
                    "Failed to delete file from the filesystem: {} error: {:?}",
                    &fname,
                    err
                );
            });

            tracing::info!("File deleted: {}", &fname);

            // Respond to the client.
            if upload_by_web {
                let template = TemplateStruct {
                    domain: String::from(headers[HOST].to_str().unwrap()),
                    message1: String::from("File Deleted"),
                    message2: String::from("woah where'd the file go???"),
                    ..Default::default()
                };
                render_template(res, &headers, "error.html", template, StatusCode::OK).await
            } else {
                res.set_status_code(StatusCode::OK);
                res.render(Text::Json(r#"{"deletion": "success"}"#));
            }
        }

        // adminkey didn't return a filename
        Err(_) => {
            // Respond to the client
            if upload_by_web {
                let template = TemplateStruct {
                    domain: String::from(headers[HOST].to_str().unwrap()),
                    message1: String::from("Error 401: Unauthorised"),
                    message2: String::from(
                        "You have entered an invalid adminkey, please enter a valid adminkey",
                    ),
                    ..Default::default()
                };
                render_template(
                    res,
                    &headers,
                    "error.html",
                    template,
                    StatusCode::UNAUTHORIZED,
                )
                .await
            } else {
                res.set_status_code(StatusCode::UNAUTHORIZED);
                res.render(Text::Json(r#"{"error": "InvalidAdminKey"}"#));
            }
        }
    };
}
