use std::path::PathBuf;

use ramhorns::{Content, Ramhorns};
use salvo::{
    addr::SocketAddr,
    hyper::{header::HOST, HeaderMap},
    prelude::StatusCode,
    writer::Text,
    Response,
};
use chrono::{DateTime, Utc};
use chrono_humanize::Humanize;

use crate::CONFIG;

// Submodules for each request handler
pub mod delete_file;
pub mod index;
pub mod list_files;
pub mod serve_file;
pub mod serve_metrics;
pub mod serve_static;
pub mod upload;

/// This file contains sub-modules that are the handlers for each type of web request.
/// Thanks audron for making me split my work up into multiple files.
/// like seriously this /is/ more clean
/// https://stackoverflow.com/questions/26435102/in-rust-what-is-the-purpose-of-a-mod-rs-file

// This is needed for templating, all the 'variables' go here!
#[derive(Content, Default)]
pub struct TemplateStruct {
    domain: String,
    fileurl: String,
    adminurl: String,
    message1: String,
    message2: String,
}

// Import tests for handlers
#[cfg(test)]
mod mod_tests;

/// This function returns what the best guess of the Client's IP address is.
///
///
/// This isn't perfect, it won't magically figure out VPN's or proxies,
/// so don't use this for authentication or anything.
///
/// # Examples
///
/// ```
/// // Clone out the header and remote_addr so we can guess the ip later
/// let headers = &req.headers().clone();
/// let remote_addr = &req.remote_addr().unwrap().clone();
/// let ip = engine::guess_ip(headers, remote_addr);
/// ```
pub fn guess_ip(headers: &HeaderMap, ip: &SocketAddr) -> String {
    // Define cloudflare ip, or None
    let cloudflare_ip: Option<String> = headers
        .get("CF-Connecting-IP")
        .map(|hv| hv.to_str().unwrap().to_owned());
    tracing::debug!("guess_ip(CF-Connecting-IP: {:?})", cloudflare_ip);
    // Define x_real_ip, or None
    let x_real_ip: Option<String> = headers
        .get("X-Real-IP")
        .map(|hv| hv.to_str().unwrap().to_owned());
    tracing::debug!("guess_ip(X-Real-IP: {:?})", x_real_ip);

    if cloudflare_ip.is_none() {
        // Either there was a problem with the cloudflare header, or there is no cloudflare.
        tracing::debug!(
            "guess_ip(No cloudflare header detected!, falling back to X-Real-IP header)"
        );
        if x_real_ip.is_none() {
            tracing::debug!("guess_ip(No X-Real-IP detected!, falling back to remote_addr header)");

            match ip {
                SocketAddr::IPv4(addr) => return addr.ip().to_string(),
                SocketAddr::IPv6(addr) => return addr.ip().to_string(),
                SocketAddr::Unix(_) => return "0.0.0.0".to_string(),
                _ => return "0.0.0.0".to_string(),
            }
        } else {
            return x_real_ip.unwrap_or("0.0.0.0".to_string());
        }
    } else {
        return cloudflare_ip.unwrap_or("0.0.0.0".to_string());
    }

    // Unable to determine ip, use 0.0.0.0
    tracing::debug!("guess_ip(headers): {:?}", headers);
    tracing::error!("guess_ip(Failed to guess ip, falling back to 0.0.0.0)");
    "0.0.0.0".to_string()
}

// Template rendering function
// When we want to render a page, we just pass over the TemplateStruct and anything else.

// I'm thinking of improving this a bit, where /all/ rendering is passed through this.
// All error handling and all branches of potential WEB and API passes through here.
// This function should be self-contained, so all it should receive is the following
// res, for building the response, and adding to any previous changes
// headers, for grabbing the domain and building the template path.
// template_filename, to tell the renderer what 'template' to render.
// TemplateStruct, which is a custom struct that contains all the variables that end up on the webpage.
pub async fn render_template(
    res: &mut Response,
    headers: &HeaderMap,
    template_filename: &str,
    template: TemplateStruct,
    status_code: StatusCode,
) {
    // Parse the host from the headers
    let host = headers
        .get(HOST)
        .and_then(|h| h.to_str().ok())
        .unwrap_or("localhost:8282");

    // Build the template path from the host
    let template_host_path = PathBuf::from("./templates/").join(host);

    // Load the template
    let tpls: Ramhorns = match Ramhorns::from_folder(template_host_path) {
        Ok(tpls) => tpls,
        Err(_) => {
            // If the template is not found, fall back to the default template file
            let default_template_host_path = PathBuf::from("./templates/localhost:8282/");
            Ramhorns::from_folder(default_template_host_path).expect("Couldn't find fallback template directory. Please ensure that the './template/localhost:8282' directory exists and is accessible.")
        }
    };

    // Tell the engine to render the template assigned above
    let rendered = tpls.get(template_filename).unwrap().render(&template);

    // Set the status code and render the completed page
    res.set_status_code(status_code);
    res.render(Text::Html(rendered));
}

/// This checks if the given mimetype is allowed or not.
/// 1 = banned, 0 = allowed
async fn is_mimetype_banned<S>(mimetype: S) -> bool
where
    S: AsRef<str>,
{
    // Load the banned files from the config
    let banned = CONFIG
        .get_array("operations.banned_mimetype")
        .expect("Couldn't find 'banned_mimetype' in config. :(");

    // Convert the provided mimetype to lowercase for case-insensitive comparison
    let lowercase_mimetype = mimetype.as_ref().to_lowercase();

    // Convert the banned mimetypes to lowercase and collect into a Vec
    let lowercase_banned: Vec<String> = banned
        .iter()
        .map(|mime| {
            let mime_str = mime
                .clone()
                .into_string()
                .expect("Invalid banned mimetype in config")
                .to_lowercase();
            tracing::debug!("Banned mimetype: {}", mime_str);
            mime_str
        })
        .collect();

    tracing::debug!("Provided mimetype: {}", lowercase_mimetype);

    lowercase_banned.contains(&lowercase_mimetype)
}


/// Convert the filesize to a human readable number
fn convert_file_size(file_size: i64) -> String {
    let units = ["B", "KB", "MB", "GB", "TB"];
    let base = 1024; // Base for conversion (1024 bytes = 1 kilobyte)

    if file_size < base {
        return format!("{} {}", file_size, units[0]);
    }

    let exponent = (file_size as f64).log(base as f64).floor() as u32;
    let converted_size = (file_size as f64 / base.pow(exponent) as f64).round();

    format!("{:.2} {}", converted_size, units[exponent as usize])
}

/// Convert unix timestamp to a human friendly timestamp
fn convert_unix_timestamp(unix_timestamp: i64) -> String {
    let datetime: DateTime<Utc> = DateTime::from_utc(
        chrono::NaiveDateTime::from_timestamp_opt(unix_timestamp, 0).unwrap(),
        Utc,
    );

    let relative_date = datetime.humanize();

    relative_date.to_string()
}