#[cfg(test)]
mod tests {
    use crate::handlers::{guess_ip, is_mimetype_banned};
    use config::Config;
    use salvo::addr::SocketAddr as SalvoSocketAddr;
    use salvo::hyper::HeaderMap;
    use std::net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr};

    // Test IP Guessing with no headers at all
    // This should end up guessing the ip from remote_addr in the request (sent through via 'ip' var)
    #[tokio::test]
    async fn test_guess_ip_no_headers() {
        let headers = HeaderMap::new();
        let ip_addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
        let std_socket_addr = SocketAddr::new(ip_addr, 8080);
        let salvo_socket_addr = SalvoSocketAddr::from(std_socket_addr);
        let guessed_ip = guess_ip(&headers, &salvo_socket_addr);
        assert_eq!(guessed_ip, "127.0.0.1");
    }

    #[tokio::test]
    async fn test_guess_ip_cf_connecting_ip() {
        let mut headers = HeaderMap::new();
        headers.insert("CF-Connecting-IP", "192.168.0.1".parse().unwrap());
        let ip_addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
        let std_socket_addr = SocketAddr::new(ip_addr, 8000);
        let salvo_socket_addr = SalvoSocketAddr::from(std_socket_addr);
        let guessed_ip = guess_ip(&headers, &salvo_socket_addr);
        assert_eq!(guessed_ip, "192.168.0.1");
    }

    #[tokio::test]
    async fn test_guess_ip_x_real_ip() {
        let mut headers = HeaderMap::new();
        headers.insert("X-Real-IP", "192.168.0.1".parse().unwrap());
        let ip_addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
        let std_socket_addr = SocketAddr::new(ip_addr, 8000);
        let salvo_socket_addr = SalvoSocketAddr::from(std_socket_addr);
        let guessed_ip = guess_ip(&headers, &salvo_socket_addr);
        assert_eq!(guessed_ip, "192.168.0.1");
    }

    #[tokio::test]
    async fn test_guess_ip_ip_v4() {
        let headers = HeaderMap::new();
        let ip_addr = IpAddr::V4(Ipv4Addr::new(192, 168, 0, 1));
        let std_socket_addr = SocketAddr::new(ip_addr, 8000);
        let salvo_socket_addr = SalvoSocketAddr::from(std_socket_addr);
        let guessed_ip = guess_ip(&headers, &salvo_socket_addr);
        assert_eq!(guessed_ip, "192.168.0.1");
    }

    #[tokio::test]
    async fn test_guess_ip_ip_v6() {
        let headers = HeaderMap::new();
        let ip_addr = IpAddr::V6(Ipv6Addr::new(0x2001, 0xdb8, 0, 0, 0, 0, 0, 1));
        let std_socket_addr = SocketAddr::new(ip_addr, 8000);
        let salvo_socket_addr = SalvoSocketAddr::from(std_socket_addr);
        let guessed_ip = guess_ip(&headers, &salvo_socket_addr);
        assert_eq!(guessed_ip, "2001:db8::1");
    }

    // Testing banned files!

    // Mock implementation of `Config`
    struct MockConfig;
    impl MockConfig {
        fn get_array(&self, key: &str) -> Option<Vec<String>> {
            match key {
                "operations.banned_mimetype" => Some(vec![
                    String::from("application/x-msdownload"),
                    String::from("application/x-msi"),
                    String::from("application/x-silverlight"),
                ]),
                _ => None,
            }
        }
    }

    #[tokio::test]
    async fn test_is_mimetype_banned() {
        // Test when mimetype is banned
        let result = is_mimetype_banned("application/x-msdownload").await;
        assert!(result);

        // Test the 3rd mimetype in the list, hopefully to catch any errors with the loop
        let result = is_mimetype_banned("application/x-silverlight").await;
        assert!(result);

        // Test when mimetype is not banned
        let result = is_mimetype_banned("audio/mp3").await;
        assert!(!result);
    }

    #[tokio::test]
    async fn test_is_mimetype_banned_case_insensitive() {
        // Test when mimetype is banned (case-insensitive)
        let result = is_mimetype_banned("APPLICATION/x-msi").await;
        assert!(result);

        // Test when mimetype is not banned (case-insensitive)
        let result = is_mimetype_banned("aUdIo/MP4").await;
        assert!(!result);
    }
}
