use core::fmt;
use salvo::{handler, writer::Text, Response};
use std::time::Instant;

use crate::{db, SQLITE};

// Define the custom data type for mimetype metrics
struct MimetypeMetric {
    host: String,
    mimetype: String,
    alive_files: i64,
    dead_files: i64,
    filesize_alive: i64,
    filesize_dead: i64,
    views_alive: i64,
    views_dead: i64,
}

// Implement the Display trait for MimetypeMetric
impl fmt::Display for MimetypeMetric {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "total_alive{{host=\"{}\", mimetype=\"{}\"}} {}\n\
            total_dead{{host=\"{}\", mimetype=\"{}\"}} {}\n\
            filesize_alive{{host=\"{}\", mimetype=\"{}\"}} {}\n\
            filesize_dead{{host=\"{}\", mimetype=\"{}\"}} {}\n\
            views_alive{{host=\"{}\", mimetype=\"{}\"}} {}\n\
            views_dead{{host=\"{}\", mimetype=\"{}\"}} {}\n\n",
            self.host,
            self.mimetype,
            self.alive_files,
            self.host,
            self.mimetype,
            self.dead_files,
            self.host,
            self.mimetype,
            self.filesize_alive,
            self.host,
            self.mimetype,
            self.filesize_dead,
            self.host,
            self.mimetype,
            self.views_alive,
            self.host,
            self.mimetype,
            self.views_dead,
        )
    }
}

// This handles the generation of the metrics page.
#[handler]
pub async fn serve_metrics(res: &mut Response) {
    // Lets start timing this:
    let start = Instant::now();

    // Setup db pool
    let sqlconn = SQLITE.get().unwrap();
    // Setup the massive string of metrics
    let mut rendered = String::new();

    ////
    // General Stats
    ////

    // Counting how many files each IP has uploaded.
    let mut ipcount = db::get_total_uploads_ip(sqlconn).await;
    // Loop through each IP and render it as a new line with a label for each IP
    for ipc in ipcount.as_mut().unwrap() {
        rendered = format!(
            "{}filecount_ip{{ip=\"{}\"}} {}\n",
            rendered,
            &ipc.0.to_string(),
            &ipc.1.to_string()
        );
    }

    // Total number of files alive/dead
    rendered = format!(
        "{}filesize_alive {}\n",
        rendered,
        db::total_alive_filesize(sqlconn).await.unwrap()
    );
    rendered = format!(
        "{}filesize_dead {}\n",
        rendered,
        db::total_dead_filesize(sqlconn).await.unwrap()
    );

    ////
    // Filetype Stats
    ////

    let mimetype_metrics = db::get_mimetype_metrics(sqlconn).await;
    // Add a newline here so we can keep it pretty.
    rendered = format!("{}\n", rendered);

    // The previous for loop can be updated to use the new MimetypeMetric struct and Display implementation like so:
    for row in mimetype_metrics.unwrap() {
        let metric = MimetypeMetric {
            host: row.host,
            mimetype: row.mimetype,
            alive_files: row.alive_files,
            dead_files: row.dead_files,
            filesize_alive: row.filesize_alive,
            filesize_dead: row.filesize_dead,
            views_alive: row.views_alive,
            views_dead: row.views_dead,
        };

        rendered = format!("{}{}", rendered, metric);
    }

    ////
    // Individual Stats
    ////

    // This is a pain, we're grabbing the individual file stats and parsing them for each file.
    let filevec = db::get_file_metrics(sqlconn).await;
    // Add a newline here so we can keep it pretty.
    rendered = format!("{}\n", rendered);
    // For each line in the Vec.
    for file in filevec.unwrap() {
        // View count per file
        rendered = format!(
            "{}file_views{{filename=\"{}\", mimetype=\"{}\"}} {}\n",
            rendered, file.filename, file.mimetype, file.views,
        );

        // Size per file
        rendered = format!(
            "{}file_size{{filename=\"{}\", mimetype=\"{}\"}} {}\n",
            rendered, file.filename, file.mimetype, file.filesize,
        );

        // Bandwidth per file
        // Views * filesize
        let bandwidth = file.views * file.filesize;
        rendered = format!(
            "{}file_bandwidth{{filename=\"{}\", mimetype=\"{}\"}} {}\n\n",
            rendered, file.filename, file.mimetype, bandwidth,
        );
    }

    //

    ////
    // Error Stats
    ////

    // Number of files blocked per filetype

    // Number of files blocked by IP?

    // Number of backend errors (by error type) (plus collision errors)

    ////
    // QR Scanning Stats
    ////

    // TODO
    // Add how long it took to get all of those metrics to the page!
    let end = Instant::now();
    rendered = format!(
        "{}\nrender_time {}\n",
        rendered,
        end.duration_since(start).as_nanos().to_owned()
    );

    // Actually render the final metrics page
    res.render(Text::Plain(rendered));
}
