use once_cell::sync::OnceCell;
use salvo::prelude::*;
use salvo::serve_static::{StaticDir, StaticFile};
use sqlx::SqlitePool;

use config::Config;
use lazy_static::lazy_static;
use std::fs::create_dir_all;
use std::time::Duration;
use tokio::{task, time};
use tracing_subscriber::filter::EnvFilter;
use tracing_subscriber::fmt;
use tracing_subscriber::prelude::*;

use crate::db::DatabaseType;

// Import sub-modules.
mod db;
mod engine;

// Import tests for main, db and engine.
mod db_tests;
#[cfg(test)]
mod engine_tests;

// Import the web request handlers modules.
mod handlers;

// I /don't think/ there's any harm to adding multiple of thse for each sql backend.
// Setup the global sqlite db
static SQLITE: OnceCell<SqlitePool> = OnceCell::new();

// Setup the config globally because I can't figure out how to pass it to functions I don't call directly.
// This is evaluated at runtime, and not compilation. \o/
lazy_static! {
    pub static ref CONFIG: Config = Config::builder()
        .add_source(config::File::with_name("config.toml"))
        .build()
        .unwrap();
}

#[tokio::main]
async fn main() {
    // Read log level from config
    let log_config = CONFIG
        .get_string("logging.level")
        .expect("Couldn't find 'sql_backend' in config. :(");

    // Construct tracing_subscriber with the filter from config.
    tracing_subscriber::registry()
        .with(fmt::layer().compact())
        .with(EnvFilter::new(log_config))
        .init();

    // Determine what SQL engine we're running.
    let sql_backend = CONFIG
        .get_string("database.sql_backend")
        .expect("Couldn't find 'sql_backend' in config. :(");

    let db_url = CONFIG
        .get_string("database.url")
        .expect("Couldn't find 'url' in config. :(");

    // Match on the database type
    match DatabaseType::from_str(&sql_backend) {
        Some(DatabaseType::Sqlite) => {
            tracing::info!("Using SQLite backend");
            let pool = SqlitePool::connect(&db_url).await.unwrap();
            SQLITE.set(pool).unwrap();
        }
        Some(DatabaseType::Postgres) => {
            tracing::info!("Using Postgres backend");
            tracing::error!("https://github.com/launchbadge/sqlx/issues/964");
            todo!("Postgres not implemented yet")
            // let pool = PgPool::connect(&psql).await;
            // TODO: figure out how to use sqlx::Pool struct https://docs.rs/sqlx/latest/sqlx/struct.Pool.html
        }
        Some(DatabaseType::Mysql) => {
            tracing::info!("Using MySQL backend");
            tracing::error!("https://github.com/launchbadge/sqlx/issues/964");
            todo!("MySQL not implemented yet")
        }
        Some(DatabaseType::Mssql) => {
            tracing::info!("Using MSSQL backend");
            tracing::error!("https://github.com/launchbadge/sqlx/issues/964");
            unimplemented!("MSSQL not implemented yet, and honestly my hatred for microsoft will probably keep it this way")
        }
        Some(DatabaseType::Oraclecloud) => {
            tracing::info!("Using OracleCloud backend");
            tracing::error!("https://github.com/launchbadge/sqlx/issues/964");
            todo!("OracleCloud not implemented yet")
        }
        _ => panic!("Unsupported SQL backend: {}", sql_backend),
    };

    // Initialise the cleaner task
    let interval = CONFIG
        .get_int("operations.cleaner_interval")
        .expect("Couldn't find 'cleaner_interval' in config. :(");
    tracing::debug!("cleaner_interval: {}", interval);
    cleaner_thread(
        interval
            .try_into()
            .expect("Cleaner interval was too long to fit in a i32.... wow"),
    );

    // Attempt to create the files directory
    create_dir_all("files").unwrap();

    // Internal routing
    let router = Router::new()
        // Main / functions
        .hoop(Logger)
        .get(handlers::index::index)
        .post(handlers::upload::upload)
        // Static Pages
        .push(Router::with_path("/services").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/about").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/czb").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/qr").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/faq").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/dmca").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/welcome").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/metrics").get(handlers::serve_metrics::serve_metrics))
        .push(Router::with_path("/favicon.ico").get(StaticFile::new("static/favicon32.webp")))
        // Static File Serving
        .push(
            Router::with_path("static/<**path>").get(StaticDir::new("static/").with_listing(true)),
        )
        // Deletion API
        .push(
            Router::with_path("/delete/<adminkey>")
                .delete(handlers::delete_file::delete_file)
                .get(handlers::delete_file::delete_file),
        )
        .push(
            Router::with_path("/edit/<adminkey>")
                .delete(handlers::delete_file::delete_file)
                .get(handlers::delete_file::delete_file),
        )
        // List of files the requester IP has uploaded.
        .push(Router::with_path("/my_files").get(handlers::list_files::list_files))
        // Serving uploaded files, or 404.
        .push(Router::with_path("<file>").get(handlers::serve_file::serve_file));

    // Read environment variables for host and port
    let host = CONFIG
        .get_string("server.host")
        .expect("Couldn't find 'host' in config. :(");
    let port = CONFIG
        .get_int("server.port")
        .expect("Couldn't find 'port' in config. :(");
    let server_url = format!("{}:{}", host, port);
    tracing::info!("Listening on http://{}", server_url);
    Server::new(TcpListener::bind(&server_url))
        .serve(router)
        .await;

    // Close SQLite before closing
    tracing::info!("Attempting to safely shut down Ephemeral.");
    SQLITE.get().expect("Problem while safely closing the database :(").close().await;
}

// This spawns a tokio task to run a interval timer forever.
// the interval timer runs every 'period' seconds.
pub fn cleaner_thread(period: i32) {
    let _forever = task::spawn(async move {
        let mut interval = time::interval(Duration::from_secs(period.try_into().unwrap()));
        let sqlconn = SQLITE.get().unwrap();
        loop {
            // Wait for the next interval
            interval.tick().await;
            // Get a vec of files that will expire this loop.
            // Also I'm sorta just realising how annoying it is to pass sqlconn through like 4 functions deep just to do something.
            let old_files_result = db::get_old_files(sqlconn).await;

            match old_files_result {
                Ok(file_list) => {
                    tracing::info!("Running Cleaner");
                    for file in file_list {
                        // Delete the file from the database
                        db::delete_file(sqlconn, &file).await.unwrap_or_else(|err| {
                            tracing::error!(
                                "Failed to delete file from database: {}, error: {:?}",
                                &file,
                                err
                            );
                            0
                        });
                        // Delete the file from the filesystem
                        engine::delete_file(&file).await.unwrap_or_else(|err| {
                            tracing::error!("Failed to delete file from database: {:?}", err);
                        });
                    }
                    tracing::info!("Cleaner finished");
                }
                Err(err) => {
                    tracing::error!("Error getting files to expire: {}", err);
                    // Return a empty Vec so it doesn't delete anything
                    return Vec::<String>::new();
                }
            }
        }
    });
}
